@@ -11,6 +11,7 @@
 # under the License.
 
 import datetime
+import urllib
 import uuid
 
 try:
@@ -18,8 +19,6 @@ try:
 except ImportError:
     etree = None
 
-from six.moves import urllib
-
 from keystoneauth1 import access
 from keystoneauth1 import exceptions
 from keystoneauth1.extras._saml2.v3 import base
