@@ -12,8 +12,6 @@
 
 import abc
 
-import six
-
 from keystoneauth1 import _utils as utils
 from keystoneauth1 import access
 from keystoneauth1 import exceptions
@@ -22,8 +20,7 @@ from keystoneauth1.identity import base
 _logger = utils.get_logger(__name__)
 
 
-@six.add_metaclass(abc.ABCMeta)
-class Auth(base.BaseIdentityPlugin):
+class Auth(base.BaseIdentityPlugin, metaclass=abc.ABCMeta):
     """Identity V2 Authentication Plugin.
 
     :param string auth_url: Identity service endpoint for authorization.
