@@ -135,7 +135,42 @@ class Adapter(base.BaseLoader):
                                 'intended to be used as the upper bound of a '
                                 'range with min_version. Mutually exclusive '
                                 'with version.'),
+                cfg.IntOpt('connect-retries',
+                           deprecated_opts=deprecated_opts.get(
+                               'connect-retries'),
+                           help='The maximum number of retries that should be '
+                                'attempted for connection errors.'),
+                cfg.FloatOpt('connect-retry-delay',
+                             deprecated_opts=deprecated_opts.get(
+                                 'connect-retry-delay'),
+                             help='Delay (in seconds) between two retries '
+                                  'for connection errors. If not set, '
+                                  'exponential retry starting with 0.5 '
+                                  'seconds up to a maximum of 60 seconds '
+                                  'is used.'),
+                cfg.IntOpt('status-code-retries',
+                           deprecated_opts=deprecated_opts.get(
+                               'status-code-retries'),
+                           help='The maximum number of retries that should be '
+                                'attempted for retriable HTTP status codes.'),
+                cfg.FloatOpt('status-code-retry-delay',
+                             deprecated_opts=deprecated_opts.get(
+                                 'status-code-retry-delay'),
+                             help='Delay (in seconds) between two retries '
+                                  'for retriable status codes. If not set, '
+                                  'exponential retry starting with 0.5 '
+                                  'seconds up to a maximum of 60 seconds '
+                                  'is used.'),
+                cfg.ListOpt('retriable-status-codes',
+                            deprecated_opts=deprecated_opts.get(
+                                'retriable-status-codes'),
+                            item_type=cfg.types.Integer(),
+                            help='List of retriable HTTP status codes that '
+                                 'should be retried. If not set default to '
+                                 ' [503]'
+                            ),
                 ]
+
         if include_deprecated:
             opts += [
                 cfg.StrOpt('interface',
@@ -154,29 +189,33 @@ class Adapter(base.BaseLoader):
         """Register the oslo_config options that are needed for an Adapter.
 
         The options that are set are:
-            :service_type:      The default service_type for URL discovery.
-            :service_name:      The default service_name for URL discovery.
-            :interface:         The default interface for URL discovery.
-                                (deprecated)
-            :valid_interfaces:  List of acceptable interfaces for URL
-                                discovery. Can be a list of any of
-                                'public', 'internal' or 'admin'.
-            :region_name:       The default region_name for URL discovery.
-            :endpoint_override: Always use this endpoint URL for requests
-                                for this client.
-            :version:           The minimum version restricted to a given Major
-                                API. Mutually exclusive with min_version and
-                                max_version.
-            :min_version:       The minimum major version of a given API,
-                                intended to be used as the lower bound of a
-                                range with max_version. Mutually exclusive with
-                                version. If min_version is given with no
-                                max_version it is as if max version is
-                                'latest'.
-            :max_version:       The maximum major version of a given API,
-                                intended to be used as the upper bound of a
-                                range with min_version. Mutually exclusive with
-                                version.
+            :service_type:        The default service_type for URL discovery.
+            :service_name:        The default service_name for URL discovery.
+            :interface:           The default interface for URL discovery.
+                                  (deprecated)
+            :valid_interfaces:    List of acceptable interfaces for URL
+                                  discovery. Can be a list of any of
+                                  'public', 'internal' or 'admin'.
+            :region_name:         The default region_name for URL discovery.
+            :endpoint_override:   Always use this endpoint URL for requests
+                                  for this client.
+            :version:             The minimum version restricted to a given
+                                  Major API. Mutually exclusive with
+                                  min_version and max_version.
+            :min_version:         The minimum major version of a given API,
+                                  intended to be used as the lower bound of a
+                                  range with max_version. Mutually exclusive
+                                  with version. If min_version is given with no
+                                  max_version it is as if max version is
+                                  'latest'.
+            :max_version:         The maximum major version of a given API,
+                                  intended to be used as the upper bound of a
+                                  range with min_version. Mutually exclusive
+                                  with version.
+            :connect_retries:     The maximum number of retries that should be
+                                  attempted for connection errors.
+            :status_code_retries: The maximum number of retries that should be
+                                  attempted for retriable HTTP status codes.
 
         :param oslo_config.Cfg conf: config object to register with.
         :param string group: The ini group to register options in.
@@ -216,36 +255,54 @@ class Adapter(base.BaseLoader):
         :rtype: :py:class:`.Adapter`
         """
         c = conf[group]
-
-        if c.valid_interfaces and hasattr(c, 'interface') and c.interface:
-            raise TypeError("interface and valid_interfaces are mutually"
-                            " exclusive. Please use valid_interfaces.")
-        if c.valid_interfaces:
-            for iface in c.valid_interfaces:
-                if iface not in ('public', 'internal', 'admin'):
-                    raise TypeError("'{iface}' is not a valid value for"
-                                    " valid_interfaces. Valid valies are"
-                                    " public, internal or admin".format(
-                                        iface=iface))
-            kwargs.setdefault('interface', c.valid_interfaces)
-        elif hasattr(c, 'interface'):
-            kwargs.setdefault('interface', c.interface)
-        kwargs.setdefault('service_type', c.service_type)
-        kwargs.setdefault('service_name', c.service_name)
-        kwargs.setdefault('region_name', c.region_name)
-        kwargs.setdefault('endpoint_override', c.endpoint_override)
-        kwargs.setdefault('version', c.version)
-        kwargs.setdefault('min_version', c.min_version)
-        kwargs.setdefault('max_version', c.max_version)
-        if kwargs['version'] and (
-                kwargs['max_version'] or kwargs['min_version']):
-            raise TypeError(
-                "version is mutually exclusive with min_version and"
-                " max_version")
-
+        process_conf_options(c, kwargs)
         return self.load_from_options(**kwargs)
 
 
+def process_conf_options(confgrp, kwargs):
+    """Set Adapter constructor kwargs based on conf options.
+
+    :param oslo_config.cfg.GroupAttr confgrp: Config object group containing
+            options to inspect.
+    :param dict kwargs: Keyword arguments suitable for the constructor of
+            keystoneauth1.adapter.Adapter. Will be modified by this method.
+            Values already set remain unaffected.
+    :raise TypeError: If invalid conf option values or combinations are found.
+    """
+    if confgrp.valid_interfaces and getattr(confgrp, 'interface', None):
+        raise TypeError("interface and valid_interfaces are mutually"
+                        " exclusive. Please use valid_interfaces.")
+    if confgrp.valid_interfaces:
+        for iface in confgrp.valid_interfaces:
+            if iface not in ('public', 'internal', 'admin'):
+                # TODO(efried): s/valies/values/ - are we allowed to fix this?
+                raise TypeError("'{iface}' is not a valid value for"
+                                " valid_interfaces. Valid valies are"
+                                " public, internal or admin".format(
+                                    iface=iface))
+        kwargs.setdefault('interface', confgrp.valid_interfaces)
+    elif hasattr(confgrp, 'interface'):
+        kwargs.setdefault('interface', confgrp.interface)
+    kwargs.setdefault('service_type', confgrp.service_type)
+    kwargs.setdefault('service_name', confgrp.service_name)
+    kwargs.setdefault('region_name', confgrp.region_name)
+    kwargs.setdefault('endpoint_override', confgrp.endpoint_override)
+    kwargs.setdefault('version', confgrp.version)
+    kwargs.setdefault('min_version', confgrp.min_version)
+    kwargs.setdefault('max_version', confgrp.max_version)
+    if kwargs['version'] and (
+            kwargs['max_version'] or kwargs['min_version']):
+        raise TypeError(
+            "version is mutually exclusive with min_version and"
+            " max_version")
+    kwargs.setdefault('connect_retries', confgrp.connect_retries)
+    kwargs.setdefault('connect_retry_delay', confgrp.connect_retry_delay)
+    kwargs.setdefault('status_code_retries', confgrp.status_code_retries)
+    kwargs.setdefault('status_code_retry_delay',
+                      confgrp.status_code_retry_delay)
+    kwargs.setdefault('retriable_status_codes', confgrp.retriable_status_codes)
+
+
 def register_argparse_arguments(*args, **kwargs):
     return adapter.register_adapter_argparse_arguments(*args, **kwargs)
 
