@@ -20,11 +20,10 @@ import platform
 import socket
 import sys
 import time
+import urllib
 import uuid
 
 import requests
-import six
-from six.moves import urllib
 
 import keystoneauth1
 from keystoneauth1 import _utils as utils
@@ -48,10 +47,11 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
 # NOTE(jamielennox): Clients will likely want to print more than json. Please
 # propose a patch if you have a content type you think is reasonable to print
 # here and we'll add it to the list as required.
-_LOG_CONTENT_TYPES = set(['application/json'])
+_LOG_CONTENT_TYPES = set(['application/json', 'text/plain'])
 
 _MAX_RETRY_INTERVAL = 60.0
 _EXPONENTIAL_DELAY_START = 0.5
+_RETRIABLE_STATUS_CODES = [503]
 
 # NOTE(efried): This is defined in oslo_middleware.request_id.INBOUND_HEADER,
 # but it didn't seem worth adding oslo_middleware to requirements just for that
@@ -92,18 +92,11 @@ def _sanitize_headers(headers):
     """Ensure headers are strings and not bytes."""
     str_dict = {}
     for k, v in headers.items():
-        if six.PY3:
-            # requests expects headers to be str type in python3, which means
-            # if we get a bytes we need to decode it into a str
-            k = k.decode('ASCII') if isinstance(k, six.binary_type) else k
-            if v is not None:
-                v = v.decode('ASCII') if isinstance(v, six.binary_type) else v
-        else:
-            # requests expects headers to be str type in python2, which means
-            # if we get a unicode we need to encode it to ASCII into a str
-            k = k.encode('ASCII') if isinstance(k, six.text_type) else k
-            if v is not None:
-                v = v.encode('ASCII') if isinstance(v, six.text_type) else v
+        # requests expects headers to be str type in python3, which means
+        # if we get a bytes we need to decode it into a str
+        k = k.decode('ASCII') if isinstance(k, bytes) else k
+        if v is not None:
+            v = v.decode('ASCII') if isinstance(v, bytes) else v
         str_dict[k] = v
     return str_dict
 
@@ -126,9 +119,9 @@ class _JSONEncoder(json.JSONEncoder):
         if isinstance(o, datetime.datetime):
             return o.isoformat()
         if isinstance(o, uuid.UUID):
-            return six.text_type(o)
+            return str(o)
         if netaddr and isinstance(o, netaddr.IPAddress):
-            return six.text_type(o)
+            return str(o)
 
         return super(_JSONEncoder, self).default(o)
 
@@ -161,7 +154,12 @@ def _determine_calling_package():
     # module does this but is far less efficient. Same story with the
     # frame walking below.  One could use ``inspect.stack()`` but it
     # has far more overhead.
-    mod_lookup = dict((m.__file__, n) for n, m in sys.modules.items()
+
+    # NOTE(jwysogla): According to the docs, we should always use copy(),
+    # because sys.modules can change during iteration, which results
+    # in a RuntimeError
+    # https://docs.python.org/3/library/sys.html#sys.modules
+    mod_lookup = dict((m.__file__, n) for n, m in sys.modules.copy().items()
                       if hasattr(m, '__file__'))
 
     # NOTE(shaleh): these are not useful because they hide the real
@@ -485,7 +483,7 @@ class Session(object):
         # so we need to actually check that this is False.
         if self.verify is False:
             string_parts.append('--insecure')
-        elif isinstance(self.verify, six.string_types):
+        elif isinstance(self.verify, str):
             string_parts.append('--cacert "%s"' % self.verify)
 
         if method:
@@ -509,7 +507,7 @@ class Session(object):
         if json:
             data = self._json.encode(json)
         if data:
-            if isinstance(data, six.binary_type):
+            if isinstance(data, bytes):
                 try:
                     data = data.decode("ascii")
                 except UnicodeDecodeError:
@@ -774,7 +772,8 @@ class Session(object):
         if connect_retries is None:
             connect_retries = self._connect_retries
         # HTTP 503 - Service Unavailable
-        retriable_status_codes = retriable_status_codes or [503]
+        retriable_status_codes = retriable_status_codes or \
+            _RETRIABLE_STATUS_CODES
         rate_semaphore = rate_semaphore or self._rate_semaphore
 
         headers = kwargs.setdefault('headers', dict())
@@ -1046,8 +1045,10 @@ class Session(object):
                 raise
 
             delay = next(connect_retry_delays)
-            logger.info('Failure: %(e)s. Retrying in %(delay).1fs.',
-                        {'e': e, 'delay': delay})
+            logger.warning('Failure: %(e)s. Retrying in %(delay).1fs.'
+                           '%(retries)s retries left',
+                           {'e': e, 'delay': delay,
+                            'retries': connect_retries})
             time.sleep(delay)
 
             return self._send_request(
@@ -1082,6 +1083,17 @@ class Session(object):
                 logger.warning("Failed to redirect request to %s as new "
                                "location was not provided.", resp.url)
             else:
+                # NOTE(TheJulia): Location redirects generally should have
+                # URI's to the destination.
+                # https://datatracker.ietf.org/doc/html/rfc7231#section-7.1.2
+                if 'params' in kwargs:
+                    kwargs['params'] = {}
+
+                if 'x-openstack-request-id' in resp.headers:
+                    kwargs['headers'].setdefault('x-openstack-request-id',
+                                                 resp.headers[
+                                                     'x-openstack-request-id'])
+
                 # NOTE(jamielennox): We don't keep increasing delays.
                 # This request actually worked so we can reset the delay count.
                 connect_retry_delays.reset()
@@ -1104,9 +1116,10 @@ class Session(object):
               status_code_retries > 0):
 
             delay = next(status_code_retry_delays)
-            logger.info('Retriable status code %(code)s. Retrying in '
-                        '%(delay).1fs.',
-                        {'code': resp.status_code, 'delay': delay})
+            logger.warning('Retriable status code %(code)s. Retrying in '
+                           '%(delay).1fs. %(retries)s retries left',
+                           {'code': resp.status_code, 'delay': delay,
+                            'retries': status_code_retries})
             time.sleep(delay)
 
             # NOTE(jamielennox): We don't keep increasing connection delays.
